from typing import List, Optional, Tuple

from pydantic import BaseModel, Field

from academia_mcp.settings import settings
from academia_mcp.utils import get_with_retries, post_with_retries, sanitize_output

EXA_SEARCH_URL = "https://api.exa.ai/search"
TAVILY_SEARCH_URL = "https://api.tavily.com/search"
BRAVE_SEARCH_URL = "https://api.search.brave.com/res/v1/web/search"
EXCLUDE_DOMAINS = ["chatpaper.com"]


def _parse_domains(query: str) -> Tuple[str, List[str]]:
    site_term = "site:"
    if site_term not in query:
        return query, []
    parts = query.split()
    query_parts = []
    include_domains = []
    for part in parts:
        if not part.startswith(site_term):
            query_parts.append(part)
            continue
        domain = part[len(site_term) :]
        if domain:
            include_domains.append(domain)
    query = " ".join(query_parts)
    return query, include_domains


class WebSearchEntry(BaseModel):  # type: ignore
    id: str = Field(description="ID of the search entry, usually the URL")
    title: str = Field(description="Title of the web page")
    content: str = Field(description="Content of the web page")


class WebSearchResponse(BaseModel):  # type: ignore
    results: List[WebSearchEntry] = Field(description="Results of the search")
    search_provider: str = Field(description="Provider used to get the results")


def web_search(
    query: str,
    limit: int = 20,
    provider: str = "tavily",
    include_domains: Optional[List[str]] = None,
) -> WebSearchResponse:
    """
    Search the web using Exa Search, Brave Search or Tavily and return normalized results.
    If the specified provider is not available, the function will try to use the next available provider.

    Args:
        query: The search query, required.
        limit: The maximum number of items to return. 20 by default, maximum 25.
        provider: The provider to use. "exa", "tavily" or "brave". "tavily" by default.
        include_domains: Optional list of domains to include in the search. None by default.
    """
    providers = ("tavily", "brave", "exa")
    assert provider in providers, "Error: provider must be either 'exa', 'tavily' or 'brave'"
    if include_domains:
        assert len(include_domains) > 0, "Error: include_domains should be a non-empty list"
        assert all(
            isinstance(domain, str) for domain in include_domains
        ), "Error: include_domains should be a list of strings"

    query, query_include_domains = _parse_domains(query)
    if query_include_domains:
        if include_domains:
            include_domains.extend(query_include_domains)
        else:
            include_domains = query_include_domains

    is_tavily_available = bool(settings.TAVILY_API_KEY)
    is_exa_available = bool(settings.EXA_API_KEY)
    is_brave_available = bool(settings.BRAVE_API_KEY)
    assert is_tavily_available or is_exa_available or is_brave_available
    availability = {
        "tavily": is_tavily_available,
        "brave": is_brave_available,
        "exa": is_exa_available,
    }

    if not availability[provider]:
        for p in providers:
            if availability[p]:
                provider = p
                break

    result: Optional[WebSearchResponse] = None
    if provider == "exa":
        result = exa_web_search(query, limit, include_domains=include_domains)
    elif provider == "brave":
        result = brave_web_search(query, limit)
    elif provider == "tavily":
        result = tavily_web_search(query, limit, include_domains=include_domains)
    assert result is not None, "Error: No provider was available"
    return result


def tavily_web_search(
    query: str, limit: int = 20, include_domains: Optional[List[str]] = None
) -> WebSearchResponse:
    """
    Search the web using Tavily and return results.

    Args:
        query: The search query, required.
        limit: The maximum number of items to return. 20 by default, maximum 25.
        include_domains: Optional list of domains to include in the search. None by default.
    """
    assert isinstance(query, str), "Error: Your search query must be a string"
    assert query.strip(), "Error: Your query should not be empty"
    assert isinstance(limit, int), "Error: limit should be an integer"
    assert 0 < limit <= 25, "Error: limit should be between 1 and 25"
    if include_domains:
        assert len(include_domains) > 0, "Error: include_domains should be a non-empty list"
        assert all(
            isinstance(domain, str) for domain in include_domains
        ), "Error: include_domains should be a list of strings"

    key = settings.TAVILY_API_KEY or ""
    assert key, "Error: TAVILY_API_KEY is not set and no api_key was provided"
    payload = {
        "query": query,
        "max_results": limit,
        "auto_parameters": True,
        "exclude_domains": EXCLUDE_DOMAINS,
    }
    if include_domains:
        payload["include_domains"] = include_domains
    response = post_with_retries(TAVILY_SEARCH_URL, payload, key)
    results = response.json()["results"]
    for result in results:
        content = " ".join(result["content"].split(" ")[:40])
        content = sanitize_output(content.strip("., "))
        result["content"] = content
        result.pop("raw_content", None)
        result.pop("score", None)
    entries = [
        WebSearchEntry(id=result["url"], title=result["title"], content=result["content"])
        for result in results
    ]
    return WebSearchResponse(results=entries, search_provider="tavily")


def exa_web_search(
    query: str, limit: int = 20, include_domains: Optional[List[str]] = None
) -> WebSearchResponse:
    """
    Search the web using Exa and return results.

    Args:
        query: The search query, required.
        limit: The maximum number of items to return. 20 by default, maximum 25.
        include_domains: Optional list of domains to include in the search. None by default.
    """
    assert isinstance(query, str), "Error: Your search query must be a string"
    assert query.strip(), "Error: Your query should not be empty"
    assert isinstance(limit, int), "Error: limit should be an integer"
    assert 0 < limit <= 25, "Error: limit should be between 1 and 25"
    if include_domains:
        assert len(include_domains) > 0, "Error: include_domains should be a non-empty list"
        assert all(
            isinstance(domain, str) for domain in include_domains
        ), "Error: include_domains should be a list of strings"

    key = settings.EXA_API_KEY or ""
    assert key, "Error: EXA_API_KEY is not set and no api_key was provided"
    payload = {
        "query": query,
        "type": "auto",
        "numResults": limit,
        "context": False,
        "excludeDomains": EXCLUDE_DOMAINS,
        "contents": {
            "text": False,
            "highlights": {
                "numSentences": 5,
            },
            "context": False,
        },
    }
    if include_domains:
        payload["includeDomains"] = include_domains

    response = post_with_retries(EXA_SEARCH_URL, payload, key)
    results = response.json()["results"]
    entries = []
    for result in results:
        content = " || ".join(result["highlights"])
        content = sanitize_output(content)
        entries.append(WebSearchEntry(id=result["url"], title=result["title"], content=content))
    return WebSearchResponse(results=entries, search_provider="exa")


def brave_web_search(query: str, limit: int = 20) -> WebSearchResponse:
    """
    Search the web using Brave and return results.

    Args:
        query: The search query, required.
        limit: The maximum number of items to return. 20 by default, maximum 20.
    """
    assert isinstance(query, str), "Error: Your search query must be a string"
    assert query.strip(), "Error: Your query should not be empty"
    assert isinstance(limit, int), "Error: limit should be an integer"
    assert 0 < limit <= 20, "Error: limit should be between 1 and 20"

    key = settings.BRAVE_API_KEY or ""
    assert key, "Error: BRAVE_API_KEY is not set and no api_key was provided"
    payload = {
        "q": query,
        "count": limit,
    }
    response = get_with_retries(BRAVE_SEARCH_URL, key, params=payload)
    results = response.json()["web"]["results"]
    entries = []
    for result in results:
        content = sanitize_output(result["description"])
        entries.append(WebSearchEntry(id=result["url"], title=result["title"], content=content))
    return WebSearchResponse(results=entries, search_provider="brave")



import re
from typing import Any, Dict, List, Optional

from markdownify import markdownify  # type: ignore
from pydantic import BaseModel, Field

from academia_mcp.settings import settings
from academia_mcp.utils import get_with_retries, post_with_retries, sanitize_output

EXA_CONTENTS_URL = "https://api.exa.ai/contents"
TAVILY_EXTRACT_URL = "https://api.tavily.com/extract"
AVAILABLE_PROVIDERS = ("basic", "exa", "tavily")
ERROR_MESSAGE = "Failed to get content from the page. Try to use another provider."


class VisitWebpageResponse(BaseModel):  # type: ignore
    id: str = Field(description="ID of the webpage, usually the URL")
    provider: str = Field(description="Provider used to get the content")
    text: Optional[str] = Field(description="Text content of the webpage", default=None)
    images: List[str] = Field(description="Images of the webpage", default_factory=list)
    error: Optional[str] = Field(
        description="Error message if the webpage is not found", default=None
    )


def _exa_visit_webpage(url: str) -> Dict[str, Any]:
    key = settings.EXA_API_KEY or ""
    assert key, "Error: EXA_API_KEY is not set and no api_key was provided"
    payload = {
        "urls": [url],
        "text": True,
    }
    response = post_with_retries(EXA_CONTENTS_URL, payload=payload, api_key=key)
    results = response.json()["results"]
    if not results:
        return {"error": ERROR_MESSAGE}
    return {"text": results[0]["text"]}


def _tavily_visit_webpage(url: str) -> Dict[str, Any]:
    key = settings.TAVILY_API_KEY or ""
    assert key, "Error: TAVILY_API_KEY is not set and no api_key was provided"
    payload = {
        "urls": [url],
        "extract_depth": "advanced",
        "include_images": True,
    }
    response = post_with_retries(TAVILY_EXTRACT_URL, payload=payload, api_key=key)
    results = response.json()["results"]
    if not results:
        return {"error": ERROR_MESSAGE}
    result = results[0]
    return {"text": result["raw_content"], "images": result["images"]}


def _basic_visit_webpage(url: str) -> Dict[str, Any]:
    try:
        response = get_with_retries(url)
        content_type = response.headers.get("content-type", "").lower()
        if not content_type or (
            not content_type.startswith("text/") and "html" not in content_type
        ):
            if settings.EXA_API_KEY:
                return _exa_visit_webpage(url)
            return {"error": f"Unsupported content-type: {content_type or 'unknown'}"}
        markdown_content = markdownify(response.text).strip()
        markdown_content = re.sub(r"\n{3,}", "\n\n", markdown_content)
        return {"text": markdown_content}
    except Exception as e:
        return {"error": str(e) + "\n" + ERROR_MESSAGE}


def visit_webpage(url: str, provider: str = "basic") -> VisitWebpageResponse:
    """
    Visit a webpage and return the content.
    Try to use both "tavily" and "basic" providers. They might work differently for the same URL.

    Args:
        url: The URL of the webpage to visit.
        provider: The provider to use. Available providers: "tavily" (default), "exa", or "basic".
    """
    assert (
        provider in AVAILABLE_PROVIDERS
    ), f"Invalid provider: {provider}. Available providers: {AVAILABLE_PROVIDERS}"

    if provider == "exa" and settings.EXA_API_KEY:
        result = _exa_visit_webpage(url)
    elif provider == "tavily" and settings.TAVILY_API_KEY:
        result = _tavily_visit_webpage(url)
    else:
        result = _basic_visit_webpage(url)

    result = VisitWebpageResponse(id=url, provider=provider, **result)
    if result.text:
        result.text = sanitize_output(result.text)
    if result.error:
        result.error = sanitize_output(result.error)
    return result



import json
from typing import List, Any, Dict

from pydantic import BaseModel

from academia_mcp.llm import llm_acall
from academia_mcp.utils import truncate_content
from academia_mcp.settings import settings


PROMPT = """You are a helpful assistant that answers questions about documents accurately and concisely.
Please answer the following questions based solely on the provided document.
If there is no answer in the document, output "There is no answer in the provided document".
First cite ALL relevant document fragments, then provide a final answer.
Answer all given questions one by one.
Make sure that you answer the actual questions, and not some other similar questions.

Questions:
{question}

Document:
==== BEGIN DOCUMENT ====
{document}
==== END DOCUMENT ====

Questions (repeated):
{question}

Your citations and answers:"""


class ChatMessage(BaseModel):  # type: ignore
    role: str
    content: str | List[Dict[str, Any]]


ChatMessages = List[ChatMessage]


async def document_qa(
    document: str | Dict[str, Any],
    question: str,
) -> str:
    """
    Answer a question about a document.
    Use this tool when you need to find relevant information in a big document.
    It takes a question and a document as inputs and generates an answer based on the document.

    Example:
    >>> document = "The quick brown fox jumps over the lazy dog."
    >>> answer = document_qa(question="What animal is mentioned? How many of them?", document=document)
    >>> print(answer)
    "The document mentions two animals: a fox and a dog. 2 animals."

    Returns an answer to all questions based on the document content.

    Args:
        question: Question (or questions) to be answered about the document.
        document: The full text of the document to analyze.
    """
    assert question and question.strip(), "Please provide non-empty 'question'"
    if isinstance(document, dict):
        document = json.dumps(document)
    assert document and document.strip(), "Please provide non-empty 'document'"

    question = truncate_content(question, settings.DOCUMENT_QA_QUESTION_MAX_LENGTH)
    document = truncate_content(document, settings.DOCUMENT_QA_DOCUMENT_MAX_LENGTH)

    model_name = settings.DOCUMENT_QA_MODEL_NAME
    prompt = PROMPT.format(question=question, document=document)
    content = await llm_acall(
        model_name=model_name, messages=[ChatMessage(role="user", content=prompt)]
    )
    return content.strip()



from academia_mcp.tools import hf_datasets_search


def test_hf_datasets_search_gazeta() -> None:
    response = hf_datasets_search(query="gazeta")
    assert response.results
    assert "IlyaGusev/gazeta" in str(response.results)


def test_hf_datasets_search_cifar() -> None:
    response = hf_datasets_search(query="CIFAR-10")
    assert response.results
    assert "uoft-cs/cifar10" in str(response.results)



from typing import Any, Dict, List, TypeVar

from openai import AsyncOpenAI
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from pydantic import BaseModel

from academia_mcp.settings import settings

T = TypeVar("T", bound=BaseModel)


class ChatMessage(BaseModel):  # type: ignore
    role: str
    content: str | List[Dict[str, Any]]


ChatMessages = List[ChatMessage]


async def llm_acall(model_name: str, messages: ChatMessages, **kwargs: Any) -> str:
    key = settings.OPENROUTER_API_KEY
    assert key, "Please set OPENROUTER_API_KEY in the environment variables"
    base_url = settings.BASE_URL

    client = AsyncOpenAI(base_url=base_url, api_key=key)
    response: ChatCompletionMessage = (
        (
            await client.chat.completions.create(
                model=model_name,
                messages=messages,
                **kwargs,
            )
        )
        .choices[0]
        .message
    )
    assert response.content, "Response content is None"
    return response.content


async def llm_acall_structured(
    model_name: str,
    messages: ChatMessages,
    response_format: type[T],
    num_parsing_retries: int = 3,
    **kwargs: Any,
) -> T:
    key = settings.OPENROUTER_API_KEY
    assert key, "Please set OPENROUTER_API_KEY in the environment variables"
    base_url = settings.BASE_URL

    client = AsyncOpenAI(base_url=base_url, api_key=key)
    converted_messages = [message.model_dump() for message in messages]
    for retry_index in range(num_parsing_retries):
        try:
            structured_response: T | None = (
                (
                    await client.chat.completions.parse(
                        model=model_name,
                        messages=converted_messages,
                        response_format=response_format,
                        **kwargs,
                    )
                )
                .choices[0]
                .message.parsed
            )
            assert structured_response
            break
        except Exception:
            if retry_index == num_parsing_retries - 1:
                raise
            continue
    assert structured_response
    return structured_response



import httpx

from academia_mcp.tools import speech_to_text
from academia_mcp.files import get_workspace_dir


async def test_speech_to_text_base(test_audio_url: str) -> None:
    result = await speech_to_text(test_audio_url)
    assert result is not None
    assert "dancing in the masquerade" in str(result).lower()


async def test_speech_to_text_local(test_audio_url: str) -> None:
    response = httpx.get(test_audio_url, timeout=10)
    response.raise_for_status()
    ext = test_audio_url.split(".")[-1]
    with open(get_workspace_dir() / f"audio_file.{ext}", "wb") as fp:
        fp.write(response.content)
    result = await speech_to_text(f"audio_file.{ext}")
    assert result is not None
    assert "dancing in the masquerade" in str(result).lower()



from typing import Dict, Any, List

from mcp import ClientSession, Tool
from mcp.types import CallToolResult
from mcp.client.streamable_http import streamablehttp_client

from tests.conftest import MCPServerTest


async def call_tool(mcp_server_test: MCPServerTest, tool: str, kwargs: Dict[str, Any]) -> Any:
    url = f"http://{mcp_server_test.host}:{mcp_server_test.port}/mcp"
    async with streamablehttp_client(url) as (
        read_stream,
        write_stream,
        _,
    ):
        async with ClientSession(read_stream, write_stream) as session:
            await session.initialize()
            result: CallToolResult = await session.call_tool(tool, kwargs)
            return result.structuredContent


async def fetch_tools(mcp_server_test: MCPServerTest) -> List[Tool]:
    url = f"http://{mcp_server_test.host}:{mcp_server_test.port}/mcp"
    all_tools: List[Tool] = []
    async with streamablehttp_client(url) as (
        read_stream,
        write_stream,
        _,
    ):
        async with ClientSession(read_stream, write_stream) as session:
            await session.initialize()
            tools_response = await session.list_tools()
            all_tools.extend(tools_response.tools)
    return all_tools


def test_server_run(mcp_server_test: MCPServerTest) -> None:
    assert mcp_server_test.server is not None
    assert mcp_server_test.is_running()


async def test_server_arxiv_search(mcp_server_test: MCPServerTest) -> None:
    query = 'ti:"PingPong: A Benchmark for Role-Playing Language Models"'
    result = await call_tool(mcp_server_test, "arxiv_search", {"query": query})
    assert result["results"][0]["authors"] == "Ilya Gusev"


async def test_server_tools_schemas(mcp_server_test: MCPServerTest) -> None:
    tools = {tool.name: tool for tool in await fetch_tools(mcp_server_test)}

    tool = tools["arxiv_search"]
    assert tool.outputSchema is not None
    assert "Number of results returned" in str(tool.outputSchema)
    tool = tools["arxiv_download"]
    assert tool.outputSchema is not None
    assert "Parsed references from the paper" in str(tool.outputSchema)
    tool = tools["visit_webpage"]
    assert tool.outputSchema is not None
    assert "Text content of the webpage" in str(tool.outputSchema)
    tool = tools["web_search"]
    assert tool.outputSchema is not None
    assert "Results of the search" in str(tool.outputSchema)
    tool = tools["s2_get_citations"]
    assert tool.outputSchema is not None
    assert "External IDs of the paper" in str(tool.outputSchema)
    tool = tools["s2_get_references"]
    assert tool.outputSchema is not None
    assert "External IDs of the paper" in str(tool.outputSchema)
    tool = tools["s2_get_info"]
    assert tool.outputSchema is not None
    assert "External IDs of the paper" in str(tool.outputSchema)
    tool = tools["extract_bitflip_info"]
    assert tool.outputSchema is not None
    assert "Innovative approach or solution" in str(tool.outputSchema)
    tool = tools["generate_research_proposals"]
    assert tool.outputSchema is not None
    assert "Innovative approach or solution" in str(tool.outputSchema)
    tool = tools["score_research_proposals"]
    assert tool.outputSchema is not None
    assert "ID of the proposal" in str(tool.outputSchema)



from academia_mcp.tools import arxiv_download


def test_arxiv_download() -> None:
    result = arxiv_download("2409.06820")
    assert "pingpong" in str(result).lower()


def test_arxiv_download_pdf() -> None:
    result = arxiv_download("2401.12474")
    assert "ditto" in str(result).lower()


def test_arxiv_download_check_structure() -> None:
    result = arxiv_download("2409.06820", include_references=True)
    assert result.title is not None
    assert result.abstract is not None
    assert result.toc is not None
    assert result.sections is not None
    assert result.references is not None
    assert result.original_format is not None


def test_arxiv_download_bug() -> None:
    paper = arxiv_download("2409.14913v2")
    assert "Performance improves incrementally upon frontier model releases" in str(paper)

    paper = arxiv_download("2501.08838v1")
    assert "ToMATO" in str(paper)

    paper = arxiv_download("2501.06964v1")
    assert "Patient-Centric" in str(paper)

    paper = arxiv_download("2412.08389v1")
    assert "enhance the efficacy of ESC systems" in str(paper)



from youtube_transcript_api import YouTubeTranscriptApi
from youtube_transcript_api.proxies import WebshareProxyConfig

from academia_mcp.settings import settings


def yt_transcript(video_url: str) -> str:
    """
    Tool to fetch the transcript of a YouTube video given its URL.

    Returns a transcript of the video as a single string.

    Args:
        video_url (str): YouTube video URL.
    """
    if "youtu.be/" in video_url:
        video_id = video_url.strip().split("youtu.be/")[-1]
    else:
        video_id = video_url.strip().split("v=")[-1]
    video_id = video_id.split("?")[0]
    proxy_config = None
    if settings.WEBSHARE_PROXY_USERNAME and settings.WEBSHARE_PROXY_PASSWORD:
        proxy_config = WebshareProxyConfig(
            proxy_username=settings.WEBSHARE_PROXY_USERNAME,
            proxy_password=settings.WEBSHARE_PROXY_PASSWORD,
        )
    api = YouTubeTranscriptApi(proxy_config=proxy_config)
    try:
        transcript = api.fetch(video_id)
    except Exception as e:
        return f"Error fetching transcript for video {video_url}: {e}"
    snippets = transcript.snippets
    return "\n".join([f"{int(entry.start)}: {' '.join(entry.text.split())}" for entry in snippets])



import logging
import socket
from logging.config import dictConfig
from typing import Literal, Optional

import fire  # type: ignore
import uvicorn
from mcp.server.fastmcp import FastMCP
from starlette.middleware.cors import CORSMiddleware
from uvicorn.config import LOGGING_CONFIG as UVICORN_LOGGING_CONFIG

from academia_mcp.settings import settings
from academia_mcp.tools.anthology_search import anthology_search
from academia_mcp.tools.arxiv_download import arxiv_download
from academia_mcp.tools.arxiv_search import arxiv_search
from academia_mcp.tools.bitflip import (
    extract_bitflip_info,
    generate_research_proposals,
    score_research_proposals,
)
from academia_mcp.tools.document_qa import document_qa
from academia_mcp.tools.hf_datasets_search import hf_datasets_search
from academia_mcp.tools.image_processing import describe_image, show_image
from academia_mcp.tools.latex import (
    compile_latex,
    get_latex_template,
    get_latex_templates_list,
    read_pdf,
)
from academia_mcp.tools.review import download_pdf_paper, review_pdf_paper, review_pdf_paper_by_url
from academia_mcp.tools.s2 import s2_get_citations, s2_get_info, s2_get_references, s2_search
from academia_mcp.tools.speech_to_text import speech_to_text
from academia_mcp.tools.visit_webpage import visit_webpage
from academia_mcp.tools.web_search import (
    brave_web_search,
    exa_web_search,
    tavily_web_search,
    web_search,
)
from academia_mcp.tools.yt_transcript import yt_transcript


def configure_uvicorn_style_logging(level: int = logging.INFO) -> None:
    config = {**UVICORN_LOGGING_CONFIG}
    config["disable_existing_loggers"] = False
    config["root"] = {"handlers": ["default"], "level": logging.getLevelName(level)}
    dictConfig(config)


def find_free_port() -> int:
    for port in range(5000, 6001):
        try:
            with socket.socket() as s:
                s.bind(("", port))
                return port
        except Exception:
            continue
    raise RuntimeError("No free port in range 5000-6000 found")


def create_server(
    streamable_http_path: str = "/mcp",
    mount_path: str = "/",
    stateless_http: bool = True,
    disable_web_search_tools: bool = False,
    disable_llm_tools: bool = False,
    port: Optional[int] = None,
    host: str = "0.0.0.0",
) -> FastMCP:
    server = FastMCP(
        "Academia MCP",
        stateless_http=stateless_http,
        streamable_http_path=streamable_http_path,
        mount_path=mount_path,
    )
    logger = logging.getLogger(__name__)

    server.add_tool(arxiv_search, structured_output=True)
    server.add_tool(arxiv_download, structured_output=True)
    server.add_tool(visit_webpage, structured_output=True)
    server.add_tool(s2_get_citations, structured_output=True)
    server.add_tool(s2_get_references, structured_output=True)
    server.add_tool(s2_get_info, structured_output=True)
    server.add_tool(s2_search, structured_output=True)
    server.add_tool(hf_datasets_search, structured_output=True)
    server.add_tool(anthology_search, structured_output=True)
    server.add_tool(get_latex_templates_list, structured_output=True)
    server.add_tool(get_latex_template, structured_output=True)
    server.add_tool(show_image)
    server.add_tool(yt_transcript)

    if settings.WORKSPACE_DIR:
        server.add_tool(compile_latex)
        server.add_tool(download_pdf_paper)
        server.add_tool(read_pdf)
    else:
        logger.warning(
            "WORKSPACE_DIR is not set, compile_latex/download_pdf_paper/read_pdf will not be available!"
        )

    if not disable_web_search_tools:
        if settings.TAVILY_API_KEY:
            server.add_tool(tavily_web_search, structured_output=True)
        if settings.EXA_API_KEY:
            server.add_tool(exa_web_search, structured_output=True)
        if settings.BRAVE_API_KEY:
            server.add_tool(brave_web_search, structured_output=True)
        if settings.EXA_API_KEY or settings.BRAVE_API_KEY or settings.TAVILY_API_KEY:
            server.add_tool(web_search, structured_output=True)
        else:
            logger.warning("No web search tools keys are set, web_search will not be available!")

    if not disable_llm_tools and settings.OPENROUTER_API_KEY:
        server.add_tool(extract_bitflip_info, structured_output=True)
        server.add_tool(generate_research_proposals, structured_output=True)
        server.add_tool(score_research_proposals, structured_output=True)
        server.add_tool(document_qa)
        server.add_tool(describe_image)
        if settings.WORKSPACE_DIR:
            server.add_tool(review_pdf_paper, structured_output=True)
            server.add_tool(review_pdf_paper_by_url, structured_output=True)
    else:
        logger.warning("No OpenRouter API key is set, LLM-related tools will not be available!")

    if settings.OPENAI_API_KEY:
        server.add_tool(speech_to_text)
    else:
        logger.warning("No OpenAI API key is set, speech_to_text will not be available!")

    if port is None:
        if settings.PORT is not None:
            port = int(settings.PORT)
        else:
            port = find_free_port()

    server.settings.port = port
    server.settings.host = host
    return server


def run(
    host: str = "0.0.0.0",
    port: Optional[int] = None,
    mount_path: str = "/",
    streamable_http_path: str = "/mcp",
    transport: Literal["stdio", "sse", "streamable-http"] = "streamable-http",
    disable_web_search_tools: bool = False,
    disable_llm_tools: bool = False,
) -> None:
    configure_uvicorn_style_logging()
    server = create_server(
        streamable_http_path=streamable_http_path,
        mount_path=mount_path,
        disable_web_search_tools=disable_web_search_tools,
        disable_llm_tools=disable_llm_tools,
        port=port,
        host=host,
    )

    if transport == "streamable-http":
        # Enable CORS for browser-based clients
        app = server.streamable_http_app()
        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["GET", "POST", "OPTIONS"],
            allow_headers=["*"],
            expose_headers=["mcp-session-id", "mcp-protocol-version"],
            max_age=86400,
        )
        uvicorn.run(
            app,
            host=server.settings.host,
            port=server.settings.port,
            log_level=server.settings.log_level.lower(),
        )
    else:
        server.run(transport=transport)


if __name__ == "__main__":
    fire.Fire(run)



from pathlib import Path
from io import BytesIO

import httpx
from openai import AsyncOpenAI
from academia_mcp.files import get_workspace_dir

from academia_mcp.settings import settings


async def speech_to_text(audio_path: str, provider: str = "openai") -> str:
    """
    Tool to convert speech to text using OpenAI's Whisper model.

    Returns transcribed text from the audio file.

    Args:
        audio_path (str): Path to the audio file.
        provider (str): Provider to use. Currently only "openai" is supported.
    """

    AVAILABLE_PROVIDERS = ("openai",)
    assert (
        provider in AVAILABLE_PROVIDERS
    ), f"Invalid provider: {provider}. Available providers: {AVAILABLE_PROVIDERS}"

    if audio_path.startswith("http"):
        response = httpx.get(audio_path, timeout=10)
        response.raise_for_status()
        ext = audio_path.split(".")[-1]
        audio_file = BytesIO(response.content)
        audio_file.name = f"audio_file.{ext}"
    else:
        full_audio_path = Path(audio_path)
        if not full_audio_path.exists():
            full_audio_path = Path(get_workspace_dir()) / audio_path
            assert full_audio_path.exists(), f"Audio file {audio_path} does not exist"
        audio_file = BytesIO(open(full_audio_path, "rb").read())
        audio_file.name = audio_path.split("/")[-1]

    assert provider == "openai"
    client = AsyncOpenAI(api_key=settings.OPENAI_API_KEY)
    result = await client.audio.transcriptions.create(
        model="gpt-4o-transcribe",
        file=audio_file,
        response_format="text",
    )
    return result



from academia_mcp.tools.review import review_pdf_paper_by_url


async def test_review_pdf_paper_by_url() -> None:
    review = await review_pdf_paper_by_url("https://arxiv.org/pdf/2502.01220")
    assert review
    assert review.format_issues



from academia_mcp.tools import web_search


def test_web_search_base() -> None:
    result = web_search("autoregressive models path-star graphs", limit=20)
    assert result.results


def test_web_search_exa() -> None:
    result = web_search("autoregressive models path-star graphs", provider="exa", limit=10)
    assert result.results
    assert result.search_provider == "exa"
    for entry in result.results:
        assert entry.content is not None


def test_web_search_tavily() -> None:
    result = web_search("autoregressive models path-star graphs", provider="tavily", limit=10)
    assert result.results
    assert result.search_provider == "tavily"
    for entry in result.results:
        assert entry.content is not None


def test_web_search_brave() -> None:
    result = web_search("autoregressive models path-star graphs", provider="brave", limit=10)
    assert "The Mystery of the Pathological" in str(result)
    assert result.search_provider == "brave"
    assert result.results
    for entry in result.results:
        assert entry.content is not None


def test_web_search_bug() -> None:
    results = web_search(
        '"Can Hiccup Supply Enough Fish to Maintain a Dragon\'s Diet?" University of Leicester'
    )
    assert results.results
    assert len(results.model_dump_json().splitlines()) == 1


def test_web_search_include_domains() -> None:
    results = web_search(
        "autoregressive models path-star graphs",
        include_domains=["wikipedia.org"],
    )
    assert results
    assert results.results
    assert len(results.results) > 0
    assert all("wikipedia.org" in result.id for result in results.results)


def test_web_search_include_query_domains() -> None:
    results = web_search(
        "site:wikipedia.org autoregressive models path-star graphs",
    )
    assert results.results
    assert all("wikipedia.org" in result.id for result in results.results)



import base64
import uuid
import tempfile
from io import BytesIO
from pathlib import Path
from typing import List, Dict, Any

from pydantic import BaseModel, Field

from academia_mcp.pdf import parse_pdf_file_to_images, parse_pdf_file, download_pdf
from academia_mcp.llm import llm_acall_structured, ChatMessage
from academia_mcp.files import get_workspace_dir
from academia_mcp.settings import settings


PROMPT = """
You are an expert peer reviewer for top CS/ML venues (e.g., NeurIPS/ICML/ACL).
Produce fair, strict, rigorous, reproducible reviews maximally useful to authors and ACs.
Be strict and punishing, but only for good reasons. Don't be afraid to reject a paper.
Be specific: cite paper sections/figures/tables when criticizing or praising.
Use actionable language ("Provide variance across 5 seeds on Dataset X; add leakage control Y").


# Summary
Summarize the paper and contributions in your own words (not the abstract).
Authors should agree with a well-written summary. No critique here.


# Strengths and Weaknesses
Please provide a thorough assessment of the strengths and weaknesses of the paper.
A good mental framing for strengths and weaknesses is to think of reasons you might accept or reject the paper.
Please touch on the following dimensions:

## Quality (Score 1-4: poor/fair/good/excellent)
Is the submission technically sound?
Are claims well supported (e.g., by theoretical analysis or experimental results)?
Are the methods used appropriate?
Is this a complete piece of work or work in progress?
Are the authors careful and honest about evaluating their work?

## Clarity (Score 1-4: poor/fair/good/excellent)
Is the submission clearly written?
Is it well organized? (If not, please make constructive suggestions for improving its clarity.)
Does it adequately inform the reader?
A superbly written paper provides enough information for an expert reader to reproduce its results.

## Significance (Score 1-4: poor/fair/good/excellent)
Are the results impactful for the community?
Are others (researchers or practitioners) likely to use the ideas or build on them?
Does the submission address a difficult task in a better way than previous work?
Does it advance our understanding/knowledge on the topic in a demonstrable way?
Does it provide unique data, unique conclusions about existing data, or a unique theoretical or experimental approach?

## Originality (Score 1-4: poor/fair/good/excellent)
Does the work provide new insights, deepen understanding, or highlight important properties of existing methods?
Is it clear how this work differs from previous contributions, with relevant citations provided?
Does the work introduce novel tasks or methods that advance the field?
Does this work offer a novel combination of existing techniques, and is the reasoning behind this combination well-articulated?
As the questions above indicates, originality does not necessarily require introducing an entirely new method.
Rather, a work that provides novel insights by evaluating existing methods, or demonstrates improved efficiency, fairness, etc. is also equally valuable.


# Scores
Try to be specific and detailed in your assessment.
Try not to set the same score for all the dimensions.
The scores for all dimensions should be independent of each other.
Scores should rely on strengths and weaknesses of the paper in each dimension.
If there are many substantial weaknesses, the score should be low.


# Questions
List 3-5 key actionable questions/suggestions.
Focus on points where author response could change your opinion or clarify confusion.
State clear criteria for your score changes.


# Limitations
Have the authors adequately addressed the limitations and potential negative societal impact of their work?
Please include constructive suggestions for improvement.
In general, authors should be rewarded rather than punished for being up front about the limitations of their work and any potential negative societal impact.
You are encouraged to think through whether any critical points are missing and provide these as feedback for the authors.


# Overall Score
6: Strong Accept - Flawless, groundbreaking impact, exceptional evaluation/reproducibility, no ethical issues
5: Accept - Solid, high impact (≥1 sub-area) or moderate-high (multiple areas), good-excellent evaluation/resources/reproducibility, no ethical issues
4: Borderline Accept - Solid, accept reasons outweigh reject (e.g., limited evaluation).
3: Borderline Reject - Solid, reject reasons outweigh accept.
2: Reject - Technical flaws, weak evaluation, inadequate reproducibility, incompletely addressed ethics
1: Strong Reject - Known results or unaddressed ethical issues

Don't be afraid to use 1, 2, 5, and 6.


# Confidence
5: Absolutely certain, very familiar with related work, checked math/details carefully
4: Confident but not certain, unlikely missed something
3: Fairly confident, possibly missed parts or unfamiliar with some work, details not carefully checked
2: Willing to defend but likely missed central parts, details not checked
1: Educated guess, not your area or hard to understand, details not checked


# Format issues
Find problems with the paper formatting. Report them separately.

# Result
Return the result as a JSON object in the following format:
{
    "summary": "...",
    "quality": {"strengths": ["..."], "weaknesses": ["..."], "reasoning": "...", "score": ...},
    "clarity": {"strengths": ["..."], "weaknesses": ["..."], "reasoning": "...", "score": ...},
    "significance": {"strengths": ["..."], "weaknesses": ["..."], "reasoning": "...", "score": ...},
    "originality": {"strengths": ["..."], "weaknesses": ["..."], "reasoning": "...", "score": ...},
    "questions": ["..."],
    "limitations": ["..."],
    "overall": {"reasoning": "...", "score": ...},
    "confidence": {"reasoning": "...", "score": ...},
    "format_issues": ["..."]
}

Always produce a correct JSON object.
"""


class AspectItem(BaseModel):  # type: ignore
    strengths: List[str] = Field(description="Strengths of the paper in a specific aspect")
    weaknesses: List[str] = Field(description="Weaknesses of the paper in a specific aspect")
    reasoning: str = Field(description="Reasoning about this aspect")
    score: int = Field(description="Score of this aspect")


class ReasoningItem(BaseModel):  # type: ignore
    reasoning: str = Field(description="Reasoning about this aspect")
    score: int = Field(description="Score of this aspect")


class ReviewResponse(BaseModel):  # type: ignore
    summary: str = Field(description="Summary of the paper")
    quality: AspectItem = Field(description="Quality of the paper")
    clarity: AspectItem = Field(description="Clarity of the paper")
    significance: AspectItem = Field(description="Significance of the paper")
    originality: AspectItem = Field(description="Originality of the paper")
    questions: List[str] = Field(description="Questions and suggestions for the authors")
    limitations: List[str] = Field(description="Limitations of the paper")
    overall: ReasoningItem = Field(description="Overall score and reasoning")
    confidence: ReasoningItem = Field(description="Confidence score and reasoning")
    format_issues: List[str] = Field(description="Format issues")


def _create_pdf_filename(pdf_url: str) -> str:
    if "arxiv.org/pdf" in pdf_url:
        pdf_filename = pdf_url.split("/")[-1]
    else:
        pdf_filename = str(uuid.uuid4())
    if not pdf_filename.endswith(".pdf"):
        pdf_filename += ".pdf"
    return pdf_filename


def download_pdf_paper(pdf_url: str) -> str:
    """
    Download a pdf file from a url to the workspace directory.

    Returns the path to the downloaded pdf file.

    Args:
        pdf_url: The url of the pdf file.
    """
    pdf_filename = _create_pdf_filename(pdf_url)
    pdf_path = Path(get_workspace_dir()) / pdf_filename
    download_pdf(pdf_url, pdf_path)
    return pdf_filename


async def review_pdf_paper(pdf_filename: str) -> ReviewResponse:
    """
    Review a pdf file with a paper.
    It parses the pdf file into images and then sends the images to the LLM for review.
    It can detect different issues with the paper formatting.
    Returns a proper NeurIPS-style review.

    Args:
        pdf_filename: The path to the pdf file.
    """
    pdf_filename_path = Path(pdf_filename)
    if not pdf_filename_path.exists():
        pdf_filename_path = Path(get_workspace_dir()) / pdf_filename

    images = parse_pdf_file_to_images(pdf_filename_path)
    text = "\n\n\n".join(parse_pdf_file(pdf_filename_path))
    content_parts: List[Dict[str, Any]] = [
        {
            "type": "text",
            "text": "Paper text:\n\n" + text,
        }
    ]
    for image in images:
        buffer_io = BytesIO()
        image.save(buffer_io, format="PNG")
        img_bytes = buffer_io.getvalue()
        image_base64 = base64.b64encode(img_bytes).decode("utf-8")
        image_content = {
            "type": "image_url",
            "image_url": {"url": f"data:image/png;base64,{image_base64}"},
        }
        content_parts.append(image_content)

    content_parts.append(
        {
            "type": "text",
            "text": "####\n\nInstructions:\n\n" + PROMPT,
        }
    )
    model_name = settings.REVIEW_MODEL_NAME
    messages = [
        ChatMessage(role="user", content=content_parts),
    ]
    result: ReviewResponse = await llm_acall_structured(
        model_name=model_name,
        messages=messages,
        response_format=ReviewResponse,
        max_completion_tokens=settings.REVIEW_MAX_COMPLETION_TOKENS,
        temperature=0.1,
    )
    return result


async def review_pdf_paper_by_url(pdf_url: str) -> ReviewResponse:
    """
    Review a pdf file with a paper by url.
    It downloads the pdf file and then reviews it.
    It parses the pdf file into images and then sends the images to the LLM for review.
    It can detect different issues with the paper formatting.
    Returns a proper NeurIPS-style review.

    Args:
        pdf_url: The url of the pdf file.
    """
    pdf_filename = _create_pdf_filename(pdf_url)
    with tempfile.TemporaryDirectory(prefix="temp_pdf_") as temp_dir:
        pdf_path = Path(temp_dir) / pdf_filename
        download_pdf(pdf_url, pdf_path)
        return await review_pdf_paper(str(pdf_path))



from academia_mcp.tools.bitflip import (
    extract_bitflip_info,
    generate_research_proposals,
    score_research_proposals,
)


async def test_bitflip_score_research_proposals() -> None:
    arxiv_id = "2503.07826"
    bit = (await extract_bitflip_info(arxiv_id)).bit
    assert bit
    proposals = await generate_research_proposals(bit=bit, num_proposals=2)
    assert proposals.proposals
    assert len(proposals.proposals) == 2
    assert proposals.proposals[0].flip
    assert proposals.proposals[1].flip
    scores = await score_research_proposals(proposals.proposals)
    assert scores.proposals
    assert len(scores.proposals) == 2
    assert scores.proposals[0].spark is not None
    assert scores.proposals[1].spark is not None
    assert scores.proposals[0].strengths is not None
    assert scores.proposals[1].strengths is not None
    assert scores.proposals[0].weaknesses is not None
    assert scores.proposals[1].weaknesses is not None



from pathlib import Path
from typing import Optional

from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    BASE_URL: str = "https://openrouter.ai/api/v1"

    OPENROUTER_API_KEY: str = ""
    TAVILY_API_KEY: Optional[str] = None
    EXA_API_KEY: Optional[str] = None
    BRAVE_API_KEY: Optional[str] = None
    OPENAI_API_KEY: Optional[str] = None

    REVIEW_MODEL_NAME: str = "openai/gpt-5"
    REVIEW_MAX_COMPLETION_TOKENS: int = 32768
    BITFLIP_MODEL_NAME: str = "deepseek/deepseek-chat-v3-0324"
    BITFLIP_MAX_COMPLETION_TOKENS: int = 16384
    DOCUMENT_QA_MODEL_NAME: str = "deepseek/deepseek-chat-v3-0324"
    DOCUMENT_QA_QUESTION_MAX_LENGTH: int = 10000
    DOCUMENT_QA_DOCUMENT_MAX_LENGTH: int = 200000
    DESCRIBE_IMAGE_MODEL_NAME: str = "gpt-4.1"

    WEBSHARE_PROXY_USERNAME: Optional[str] = None
    WEBSHARE_PROXY_PASSWORD: Optional[str] = None

    PORT: int = 5056
    WORKSPACE_DIR: Optional[Path] = None

    model_config = SettingsConfigDict(
        env_file=".env",
        env_file_encoding="utf-8",
        env_prefix="",
        extra="ignore",
    )


settings = Settings()



from pathlib import Path
from typing import List

import pymupdf  # type: ignore
from PIL import Image
from pypdf import PdfReader

from academia_mcp.utils import get_with_retries


def download_pdf(url: str, output_path: Path) -> None:
    response = get_with_retries(url)
    response.raise_for_status()
    content_type = response.headers.get("content-type")
    assert content_type
    assert "application/pdf" in content_type.lower()
    with open(output_path.resolve(), "wb") as fp:
        fp.write(response.content)


def parse_pdf_file(pdf_path: Path) -> List[str]:
    # Why not Marker? Because it is too heavy.
    reader = PdfReader(str(pdf_path.resolve()))

    pages = []
    for page_number, page in enumerate(reader.pages, start=1):
        try:
            text = page.extract_text()
            if not text:
                continue
            prefix = f"## Page {page_number}\n\n"
            pages.append(prefix + text)
        except Exception:
            continue
    return pages


def parse_pdf_file_to_images(pdf_path: Path) -> List[Image.Image]:
    doc = pymupdf.open(str(pdf_path.resolve()))
    images = []
    for page in doc:
        pil_image: Image.Image = page.get_pixmap().pil_image()
        images.append(pil_image)
    return images



from .anthology_search import anthology_search
from .arxiv_download import arxiv_download
from .arxiv_search import arxiv_search
from .bitflip import extract_bitflip_info, generate_research_proposals, score_research_proposals
from .document_qa import document_qa
from .hf_datasets_search import hf_datasets_search
from .image_processing import describe_image, show_image
from .latex import compile_latex, get_latex_template, get_latex_templates_list, read_pdf
from .review import download_pdf_paper, review_pdf_paper, review_pdf_paper_by_url
from .s2 import s2_get_citations, s2_get_info, s2_get_references, s2_search
from .speech_to_text import speech_to_text
from .visit_webpage import visit_webpage
from .web_search import brave_web_search, exa_web_search, tavily_web_search, web_search
from .yt_transcript import yt_transcript

__all__ = [
    "arxiv_search",
    "arxiv_download",
    "anthology_search",
    "s2_get_references",
    "s2_get_citations",
    "s2_get_info",
    "s2_search",
    "hf_datasets_search",
    "document_qa",
    "compile_latex",
    "get_latex_template",
    "get_latex_templates_list",
    "web_search",
    "tavily_web_search",
    "exa_web_search",
    "brave_web_search",
    "visit_webpage",
    "extract_bitflip_info",
    "generate_research_proposals",
    "score_research_proposals",
    "review_pdf_paper",
    "review_pdf_paper_by_url",
    "download_pdf_paper",
    "read_pdf",
    "show_image",
    "describe_image",
    "speech_to_text",
    "yt_transcript",
]



from academia_mcp.tools import yt_transcript


def test_yt_transcript_base() -> None:
    result = yt_transcript("https://www.youtube.com/watch?v=21EYKqUsPfg")
    assert result is not None
    assert "chatting with richard sutton" in result.lower()


def test_yt_transcript_short_link() -> None:
    result = yt_transcript("https://youtu.be/21EYKqUsPfg?si=iity_X55GIWUQWuT")
    assert result is not None
    assert "chatting with richard sutton" in result.lower()



# Based on
# https://api.semanticscholar.org/api-docs/graph#tag/Paper-Data/operation/get_graph_get_paper_citations

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field

from academia_mcp.utils import get_with_retries

BASE_URL = "https://api.semanticscholar.org/graph/v1"
PAPER_URL_TEMPLATE = "{base_url}/paper/{paper_id}"
CITATIONS_URL_TEMPLATE = "{base_url}/paper/{paper_id}/citations"
REFERENCES_URL_TEMPLATE = "{base_url}/paper/{paper_id}/references"
SEARCH_URL_TEMPLATE = "{base_url}/paper/search"
FIELDS = "paperId,title,authors,externalIds,venue,citationCount,publicationDate"


class S2PaperInfo(BaseModel):  # type: ignore
    arxiv_id: Optional[str] = Field(description="ArXiv ID of the paper", default=None)
    external_ids: Optional[Dict[str, Any]] = Field(
        description="External IDs of the paper.", default=None
    )
    title: str = Field(description="Paper title")
    authors: List[str] = Field(description="Authors of the paper")
    venue: str = Field(description="Paper venue")
    citation_count: Optional[int] = Field(description="Paper citation count", default=None)
    publication_date: Optional[str] = Field(description="Paper publication date", default=None)


class S2SearchResponse(BaseModel):  # type: ignore
    total_count: int = Field(description="Total number of results.")
    returned_count: int = Field(description="Number of results returned.")
    offset: int = Field(description="Offset of the results.")
    results: List[S2PaperInfo] = Field(description="Search entries")


def _format_authors(authors: List[Dict[str, Any]]) -> List[str]:
    return [a["name"] for a in authors]


def _clean_entry(entry: Dict[str, Any]) -> S2PaperInfo:
    if "citingPaper" in entry:
        entry = entry["citingPaper"]
    elif "citedPaper" in entry:
        entry = entry["citedPaper"]
    external_ids = entry.get("externalIds")
    if not external_ids:
        external_ids = dict()
    external_ids.pop("CorpusId", None)
    arxiv_id = external_ids.pop("ArXiv", None)
    return S2PaperInfo(
        arxiv_id=arxiv_id,
        external_ids=external_ids if external_ids else None,
        title=entry["title"],
        authors=_format_authors(entry["authors"]),
        venue=entry.get("venue", ""),
        citation_count=entry.get("citationCount"),
        publication_date=entry.get("publicationDate"),
    )


def _format_entries(
    entries: List[Dict[str, Any]],
    start_index: int,
    total_results: int,
) -> S2SearchResponse:
    clean_entries = [_clean_entry(e) for e in entries]
    return S2SearchResponse(
        total_count=total_results,
        returned_count=len(entries),
        offset=start_index,
        results=clean_entries,
    )


def s2_get_citations(
    arxiv_id: str,
    offset: int = 0,
    limit: int = 50,
) -> S2SearchResponse:
    """
    Get all papers that cited a given arXiv paper based on Semantic Scholar info.

    Args:
        arxiv_id: The ID of a given arXiv paper.
        offset: The offset to scroll through citations. 10 items will be skipped if offset=10. 0 by default.
        limit: The maximum number of items to return. limit=50 by default.
    """

    assert isinstance(arxiv_id, str), "Error: Your arxiv_id must be a string"
    if "v" in arxiv_id:
        arxiv_id = arxiv_id.split("v")[0]
    paper_id = f"arxiv:{arxiv_id}"

    url = CITATIONS_URL_TEMPLATE.format(base_url=BASE_URL, paper_id=paper_id)
    payload = {"fields": FIELDS, "offset": offset, "limit": limit}
    response = get_with_retries(url, params=payload)
    result = response.json()
    entries = result["data"]
    total_count = len(result["data"]) + result["offset"]

    if "next" in result:
        paper_url = PAPER_URL_TEMPLATE.format(base_url=BASE_URL, paper_id=paper_id)
        payload = {"fields": FIELDS}
        paper_response = get_with_retries(paper_url, params=payload)
        paper_result = paper_response.json()
        total_count = paper_result["citationCount"]

    return _format_entries(entries, offset if offset else 0, total_count)


def s2_get_references(
    arxiv_id: str,
    offset: int = 0,
    limit: int = 50,
) -> S2SearchResponse:
    """
    Get all papers that were cited by a given arXiv paper (references) based on Semantic Scholar info.

    Args:
        arxiv_id: The ID of a given arXiv paper.
        offset: The offset to scroll through citations. 10 items will be skipped if offset=10. 0 by default.
        limit: The maximum number of items to return. limit=50 by default.
    """
    assert isinstance(arxiv_id, str), "Error: Your arxiv_id must be a string"
    if "v" in arxiv_id:
        arxiv_id = arxiv_id.split("v")[0]
    paper_id = f"arxiv:{arxiv_id}"

    url = REFERENCES_URL_TEMPLATE.format(base_url=BASE_URL, paper_id=paper_id)
    payload = {"fields": FIELDS, "offset": offset, "limit": limit}
    response = get_with_retries(url, params=payload)
    result = response.json()
    entries = result["data"]
    total_count = len(result["data"]) + result["offset"]
    return _format_entries(entries, offset if offset else 0, total_count)


def s2_get_info(arxiv_id: str) -> S2PaperInfo:
    """
    Get the S2 info for a given arXiv ID.

    Args:
        arxiv_id: The ID of a given arXiv paper.
    """
    assert isinstance(arxiv_id, str), "Error: Your arxiv_id must be a string"
    if "v" in arxiv_id:
        arxiv_id = arxiv_id.split("v")[0]
    paper_id = f"arxiv:{arxiv_id}"
    payload = {"fields": FIELDS}
    paper_url = PAPER_URL_TEMPLATE.format(base_url=BASE_URL, paper_id=paper_id)
    response = get_with_retries(paper_url, params=payload)
    json_data = response.json()
    return S2PaperInfo(
        arxiv_id=json_data.get("externalIds", {}).get("ArXiv"),
        external_ids=json_data.get("externalIds", {}),
        title=json_data["title"],
        authors=_format_authors(json_data["authors"]),
        venue=json_data.get("venue", ""),
        citation_count=int(json_data.get("citationCount", 0)),
        publication_date=str(json_data.get("publicationDate", "")),
    )


def s2_search(
    query: str,
    offset: int = 0,
    limit: int = 5,
    min_citation_count: int = 0,
    publication_date: Optional[str] = None,
) -> S2SearchResponse:
    """
    Search the S2 corpus for a given query.

    Args:
        query: The query to search for.
        offset: The offset to scroll through results. 10 items will be skipped if offset=10. 0 by default.
        limit: The maximum number of items to return. limit=50 by default.
        min_citation_count: The minimum citation count to return. 0 by default.
        publication_date: Restricts results to the given range of publication dates or years (inclusive).
            Accepts the format <startDate>:<endDate> with each date in YYYY-MM-DD format. None by default.
    """
    url = SEARCH_URL_TEMPLATE.format(base_url=BASE_URL)
    payload = {
        "query": query,
        "offset": offset,
        "limit": limit,
        "minCitationCount": min_citation_count,
        "fields": FIELDS,
    }
    if publication_date:
        payload["publicationDateOrYear"] = publication_date
    response = get_with_retries(url, params=payload, backoff_factor=10.0, num_retries=5)
    result = response.json()
    if "data" not in result:
        return S2SearchResponse(
            total_count=0,
            returned_count=0,
            offset=offset if offset else 0,
            results=[],
        )
    entries = result["data"]
    total_count = result["total"]
    return _format_entries(entries, offset if offset else 0, total_count)



from pathlib import Path

from academia_mcp.settings import settings

DIR_PATH = Path(__file__).parent
ROOT_PATH = DIR_PATH.parent
DEFAULT_LATEX_TEMPLATES_DIR_PATH: Path = DIR_PATH / "latex_templates"


def get_workspace_dir() -> Path:
    assert settings.WORKSPACE_DIR is not None, "Please set the WORKSPACE_DIR environment variable"
    directory = Path(settings.WORKSPACE_DIR)
    if not directory.exists():
        directory.mkdir(parents=True, exist_ok=True)
    return directory



import json
import tempfile
from pathlib import Path

from academia_mcp.tools.latex import (
    compile_latex,
    get_latex_template,
    get_latex_templates_list,
    read_pdf,
)


def test_latex_get_latex_templates_list() -> None:
    templates_list = get_latex_templates_list().templates
    assert len(templates_list) > 0
    assert "agents4science_2025" in templates_list


def test_latex_get_latex_template() -> None:
    result = get_latex_template("agents4science_2025")
    assert result.template is not None
    assert result.style is not None


def test_latex_compile_latex_from_file() -> None:
    template = get_latex_template("agents4science_2025")
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        tex_filename = "temp.tex"
        tex_file_path = temp_dir_path / tex_filename
        pdf_filename = "test.pdf"
        tex_file_path.write_text(template.template, encoding="utf-8")
        result = compile_latex(str(tex_file_path), pdf_filename)
    assert "Compilation successful" in result


def test_latex_read_pdf() -> None:
    template = get_latex_template("agents4science_2025")
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        tex_filename = "temp.tex"
        tex_file_path = temp_dir_path / tex_filename
        pdf_filename = "test.pdf"
        tex_file_path.write_text(template.template, encoding="utf-8")
        compile_latex(str(tex_file_path), pdf_filename)
        read_result = json.loads(read_pdf(pdf_filename))
        assert read_result
        assert "Page 1" in read_result[0]



from io import BytesIO

import httpx
from PIL import Image

from academia_mcp.tools import show_image, describe_image
from academia_mcp.files import get_workspace_dir


def test_show_image_base(test_image_url: str) -> None:
    result = show_image(test_image_url)
    assert result is not None
    assert "image_base64" in result
    assert result["image_base64"] is not None


def test_show_image_local(test_image_url: str) -> None:
    response = httpx.get(test_image_url, timeout=10)
    response.raise_for_status()
    image = Image.open(BytesIO(response.content))
    image.save(get_workspace_dir() / "test_image.png")
    result = show_image("test_image.png")
    assert result is not None
    assert "image_base64" in result
    assert result["image_base64"] is not None


async def test_describe_image_base(test_image_url: str) -> None:
    result = await describe_image(test_image_url)
    assert result is not None
    assert "Interrogator" in result


async def test_describe_image_text(test_image_url: str) -> None:
    result = await describe_image(test_image_url, description_type="text")
    assert result is not None
    assert '"text": "Interrogator"' in result



from academia_mcp.tools import visit_webpage


def test_visit_webpage_basic() -> None:
    content = visit_webpage("https://example.com/", provider="basic")
    assert content.text is not None
    assert content.id == "https://example.com/"
    assert content.provider == "basic"
    assert "Example Domain" in content.text


def test_visit_webpage_exa() -> None:
    content = visit_webpage("https://example.com/", provider="exa")
    assert content.text is not None
    assert content.id == "https://example.com/"
    assert content.provider == "exa"
    assert "Example Domain" in content.text


def test_visit_webpage_pdf() -> None:
    content = visit_webpage("https://arxiv.org/pdf/2409.06820")
    assert content.text is not None
    assert "A Benchmark for Role-Playing" in content.text


def test_visit_webpage_nature() -> None:
    url = "https://www.nature.com/nature/articles?page=51&searchType=journalSearch&sort=PubDate&type=article&year=2020"
    content = visit_webpage(url, provider="basic")
    assert content.text is not None
    assert "1002" in content.text


def test_visit_webpage_exception() -> None:
    url = "https://www.researchgate.net/profile/Peter-Giovannini"
    content = visit_webpage(url)
    assert content.error is not None



from datetime import datetime
from typing import List, Literal, Optional

from huggingface_hub import DatasetInfo, HfApi, hf_hub_download
from pydantic import BaseModel, Field

HF_API = HfApi()


class HFDatasetsSearchEntry(BaseModel):  # type: ignore
    id: str = Field(description="ID of the dataset")
    created_at: str = Field(description="Created date of the dataset")
    last_modified: str = Field(description="Last modified date of the dataset")
    downloads: int = Field(description="Downloads of the dataset")
    likes: int = Field(description="Likes of the dataset")
    tags: List[str] = Field(description="Tags of the dataset")
    readme: str = Field(description="README of the dataset")


class HFDatasetsSearchResponse(BaseModel):  # type: ignore
    results: List[HFDatasetsSearchEntry] = Field(description="List of datasets")


def _format_date(dt: Optional[datetime]) -> str:
    if not dt:
        return ""
    return dt.strftime("%B %d, %Y")


def _clean_entry(entry: DatasetInfo) -> HFDatasetsSearchEntry:
    try:
        readme_path = hf_hub_download(repo_id=entry.id, repo_type="dataset", filename="README.md")
        with open(readme_path, "r", encoding="utf-8") as f:
            readme_content = f.read()
    except Exception:
        readme_content = ""

    return HFDatasetsSearchEntry(
        id=entry.id,
        created_at=_format_date(entry.created_at),
        last_modified=_format_date(entry.last_modified),
        downloads=entry.downloads,
        likes=entry.likes,
        tags=entry.tags,
        readme=readme_content,
    )


def _format_entries(entries: List[DatasetInfo]) -> HFDatasetsSearchResponse:
    clean_entries = [_clean_entry(entry) for entry in entries]
    return HFDatasetsSearchResponse(results=clean_entries)


def hf_datasets_search(
    query: Optional[str] = None,
    search_filter: Optional[List[str]] = None,
    limit: int = 5,
    sort_by: str = "trending_score",
    sort_order: str = "descending",
) -> HFDatasetsSearchResponse:
    """
    Search or filter HF datasets.

    Examples:
        List only the datasets in Russian for language modeling:
        hf_datasets_search(filter=(language:ru", "task_ids:language-modeling"))

        List all recent datasets with "text" in their name
        hf_datasets_search(query="text", sort_by="last_modified")

    Args:
        query: The search query for the exact match search.
        search_filter: A list of string to filter datasets.
        limit: The maximum number of items to return. limit=5 by default, limit=10 is the maximum.
        sort_by:
            The key with which to sort the resulting models.
            Possible values are "last_modified", "trending_score", "created_at", "downloads" and "likes".
            "trending_score" by default.
        sort_order: 2 sort orders: ascending, descending. descending by default.
    """
    direction: Optional[Literal[-1]] = -1 if sort_order == "descending" else None
    results = list(
        HF_API.list_datasets(
            search=query,
            sort=sort_by,
            direction=direction,
            limit=limit,
            filter=search_filter,
        )
    )
    return _format_entries(results)



# Based on
# https://arxiv.org/abs/2504.12976
# https://web.stanford.edu/class/cs197c/slides/02-literature-search.pdf

import json
import random
from typing import Any, Dict, List, Optional

from datasets import load_dataset  # type: ignore
from pydantic import BaseModel, Field

from academia_mcp.llm import ChatMessage, llm_acall_structured
from academia_mcp.settings import settings
from academia_mcp.tools.arxiv_download import arxiv_download
from academia_mcp.utils import encode_prompt


class ProposalDataset:
    dataset: Optional[List[Any]] = None

    @classmethod
    def get_dataset(cls) -> List[Any]:
        if cls.dataset is None:
            cls.dataset = list(load_dataset("UniverseTBD/hypogen-dr1")["train"])
        return cls.dataset


EXTRACT_PROMPT = """
You are a highly advanced research assistant.
You specialize in reading scientific papers for hypothesis generation and identifying innovative ideas.


## Example (BERT in NLP)
Before you begin, let 's revisit the Bit-Flip concept with an example (BERT in NLP):
- Bit: Traditional NLP models (RNNs, LSTMs) process text sequentially,
limiting their ability to understand long-range dependencies and fully capture bidirectional context.
- Flip: Instead, consider entire sentences at once, allowing context from both directions. This helps capture nuanced relationships among words.
- Spark: Bidirectional context for NLP.

## Framework
A Bit-Flip inverts a commonly held assumption,
questioning existing constraints or reapplying techniques to new domains/scales.
The "Bit" is the prevailing belief, and the "Flip" is the counterargument.

## Guidance for analysis
1. Bit (Technical Insight):
- Provide at least two sentences clearly stating the status quo or conventional approach.
- Highlight the limitation or problem it creates.
- Include enough detail so it is self-contained and does not rely on additional context from elsewhere.
2. Flip (Innovation):
- Provide at least two sentences describing the novel approach or perspective.
- Explain the method or technique that enables this change.
- Include enough detail so the Flip is understandable on its own.
3. Spark (Core Summary):
- A concise 4-6 word phrase capturing the core idea.

Now, consider this research abstract:
{{abstract}}

Your task:
Identify the Bit, Flip, and Spark from the abstract in a detailed manner:
- Bit: at least two sentences, with sufficient detail about the conventional approach and its limitation.
- Flip: at least two sentences, describing the new approach or perspective with enough detail to understand the main technique.
- Spark: a concise 4-6 word summary of the core idea.

Follow these rules:
- Do not cite the paper itself or its authors.
- Instead of saying "We/I introduced an idea", just say "An idea was introduced ...".

Return only the JSON object in this exact format (no extra text):
{
    "bit": "Technical limitation or conventional approach, in at least two sentences",
    "flip": "Innovative approach or solution, in at least two sentences",
    "spark": "4-6 word summary"
}
"""

SYSTEM_IMPROVEMENT_PROMPT = """
You are a brilliant rogue researcher who has just discovered a secret laboratory hidden behind a bookshelf.
This lab doesn't follow the boring rules of conventional academia: here, the wildest ideas become breakthrough innovations.

You've spent years watching researchers play it safe with incremental improvements and mind-numbing iterations.
But tonight you're ready to unleash scientific chaos.
"""

IMPROVEMENT_PROMPT = """
You've been handed a Bit: a technical limitation, constraint, or conventional approach from some stuffy research paper.
This Bit represents everything predictable and safe about current methods.
Your task: Shatter it completely.
Create a Flip: wildly unconventional improvement that would make traditional researchers choke on their coffee.
Then capture its essence in a Spark: brilliant summary that crackles with innovation.

- Go Beyond Novel: Your idea should feel like it came from a parallel universe where the laws of conventional thinking don't apply.
- Make It Automatically Verifiable: No human babysitters allowed. Your idea must prove itself through pure computational audacity.
- Be Surgically Specific: Vague hand-waving is for amateur rebels. Describe exactly how your mad science would work.
- Feasibility is the Key: Make it simple and executable. Try to make it reproducible.
- Channel Your Inner Contrarian: What would happen if you took the exact opposite approach?
What if you turned the problem inside-out, upside-down, or into a completely different dimension?

{% for example in examples %}
## Example {{loop.index}} of boring research
- Bit: {{example["bit"]}}
- Chain of reasoning: {{example["chain_of_reasoning"]}}
- Flip: {{example["flip"]}}
- Spark: {{example["spark"]}}
{% endfor %}

Those examples are boring, your ideas should not be like that.

Now, please propose a chain of reasoning that leads to an improvement idea for this Bit:
{{bit}}

{% if additional_context %}Additional context:
{{additional_context}}{% endif %}

Finalize your idea by providing the idea details:
- Abstract: An abstract that summarizes the proposal in conference format (approximately 250 words).
- Experiments: A list of experiments that would be conducted to validate the proposal.
Ensure these are simple and feasible. Be specific in exactly how you would test the hypothesis, and detail precise algorithmic changes.
Include the evaluation metrics you would use.
- Risks and limitations: A list of potential risks and limitations of the proposal.

Generate {{num_proposals}} proposals.

Return only the JSON list of proposals in this exact format:
{
    "proposals": [
        {
            "chain_of_reasoning": "Chain of reasoning that leads to an improvement idea for this Bit. At least 5 sentences.",
            "flip": "Your wildest, craziest, most innovative idea, in at least two sentences",
            "spark": "4-6 word summary",
            "abstract": "An abstract that summarizes the proposal in conference format (approximately 250 words).",
            "experiments": ["...", "..."],
            "risks_and_limitations": ["...", "..."]
        },
        ...
    ]
}
"""


SCORE_PROMPT = """
You are a highly advanced research assistant.
You are given a list of research proposals.
Your task is to score the proposals.

Proposals:
{% for proposal in proposals %}
----
{{proposal}}
----
{% endfor %}

Here are the criteria:
- "Strengths": A list of strengths of the proposal.
- "Weaknesses": A list of weaknesses of the proposal.
- "Novelty": Is the proposal novel? A rating from 1 to 4 (low, medium, high, very high).
- "Clarity": Is the proposal clear? A rating from 1 to 4 (low, medium, high, very high).
- "Significance": Is the proposal significant? A rating from 1 to 4 (low, medium, high, very high).
- "Feasibility": Is the proposal feasible and easy to implement? A rating from 1 to 4 (low, medium, high, very high).
- "Soundness": Is the proposal sound? A rating from 1 to 4 (poor, fair, good, excellent).
- "Overall": A rating from 1 to 10 (very strong reject to award quality).

Return only scores for all proposals in this exact format (no extra text):
{
    "proposals": [
        {
            "proposal_id": 0,
            "spark": "...",
            "strengths": ["...", "..."],
            "weaknesses": ["...", "..."],
            "novelty": 2,
            "clarity": 2,
            "significance": 2,
            "feasibility": 2,
            "soundness": 2,
            "overall": 5
        },
        ...
    ]
}
"""


class BitFlipInfo(BaseModel):  # type: ignore
    bit: str = Field(description="Technical limitation or conventional approach")
    flip: str = Field(description="Innovative approach or solution")
    spark: str = Field(description="4-6 word summary")


async def extract_bitflip_info(arxiv_id: str) -> BitFlipInfo:
    """
    Extracts the Bit-Flip information from the arXiv paper.

    A Bit-Flip is a technique that inverts a commonly held assumption,
    questioning existing constraints or reapplying techniques to new domains/scales.
    The "Bit" is the prevailing belief, and the "Flip" is the counterargument.

    Args:
        arxiv_id: The arXiv ID of the paper to extract the Bit-Flip information from.
    """
    model_name = settings.BITFLIP_MODEL_NAME
    paper = arxiv_download(arxiv_id)
    abstract = paper.abstract
    prompt = encode_prompt(EXTRACT_PROMPT, abstract=abstract)
    bitflip_info: BitFlipInfo = await llm_acall_structured(
        model_name=model_name,
        messages=[ChatMessage(role="user", content=prompt)],
        temperature=0.0,
        response_format=BitFlipInfo,
    )
    return bitflip_info


class ResearchProposal(BaseModel):  # type: ignore
    proposal_id: int = Field(default=0, description="ID of the proposal")
    flip: str = Field(description="Innovative approach or solution, in at least two sentences")
    spark: str = Field(description="4-6 word summary")
    abstract: str = Field(
        description="An abstract that summarizes the proposal in conference format."
    )
    experiments: List[str] = Field(
        description="A list of experiments that would be conducted to validate the proposal."
    )
    risks_and_limitations: List[str] = Field(
        description="A list of potential risks and limitations of the proposal."
    )


class GenerateResearchProposalResponse(BaseModel):  # type: ignore
    proposals: List[ResearchProposal] = Field(description="A list of research proposals")


async def generate_research_proposals(
    bit: str, num_proposals: int = 3, additional_context: str = ""
) -> GenerateResearchProposalResponse:
    """
    Proposes improvement ideas for the Bit.

    Args:
        bit: The Bit to propose improvement ideas for. The bit is a technical limitation or conventional approach of some paper.
        num_proposals: The number of proposals to generate.
        additional_context: Additional context to use when proposing the improvement idea.
    """
    model_name = settings.BITFLIP_MODEL_NAME
    max_completion_tokens = int(settings.BITFLIP_MAX_COMPLETION_TOKENS)
    examples = ProposalDataset.get_dataset()[:]
    examples = random.choices(examples, k=2)

    prompt = encode_prompt(
        IMPROVEMENT_PROMPT,
        bit=bit,
        examples=examples,
        num_proposals=num_proposals,
        additional_context=additional_context,
    )
    result: GenerateResearchProposalResponse = await llm_acall_structured(
        model_name=model_name,
        messages=[
            ChatMessage(role="system", content=SYSTEM_IMPROVEMENT_PROMPT),
            ChatMessage(role="user", content=prompt),
        ],
        response_format=GenerateResearchProposalResponse,
        max_completion_tokens=max_completion_tokens,
        temperature=1.0,
    )
    return result


class ScoredProposal(BaseModel):  # type: ignore
    proposal_id: int = Field(default=0, description="ID of the proposal")
    spark: str = Field(description="4-6 word summary")
    strengths: List[str] = Field(description="A list of strengths of the proposal")
    weaknesses: List[str] = Field(description="A list of weaknesses of the proposal")
    novelty: int = Field(description="Novelty rating from 1 to 4")
    clarity: int = Field(description="Clarity rating from 1 to 4")
    significance: int = Field(description="Significance rating from 1 to 4")
    feasibility: int = Field(description="Feasibility rating from 1 to 4")
    soundness: int = Field(description="Soundness rating from 1 to 4")
    overall: int = Field(description="Overall rating from 1 to 10")


class ScoreResearchProposalsResponse(BaseModel):  # type: ignore
    proposals: List[ScoredProposal] = Field(description="List of scored proposals")


async def score_research_proposals(
    proposals: str | List[Dict[str, Any]] | List[Any] | List[str] | List[ResearchProposal],
) -> ScoreResearchProposalsResponse:
    """
    Scores a list of research proposals.
    Use proposals obtained with the `generate_research_proposal` tool.

    Args:
        proposals: A list of JSON strings with research proposals.
    """
    model_name = settings.BITFLIP_MODEL_NAME
    if isinstance(proposals, str):
        proposals = json.loads(proposals)
    if isinstance(proposals, list) and proposals:
        proposals = [str(p) for p in proposals]
    assert isinstance(proposals, list), "Proposals should be a list"
    prompt = encode_prompt(SCORE_PROMPT, proposals=proposals)
    result: ScoreResearchProposalsResponse = await llm_acall_structured(
        model_name=model_name,
        messages=[ChatMessage(role="user", content=prompt)],
        response_format=ScoreResearchProposalsResponse,
        temperature=0.0,
    )
    return result



import asyncio
import base64
import contextlib
import json
import logging
import os
import threading
from io import BytesIO
from pathlib import Path
from textwrap import dedent
from typing import Dict, List, Optional, Any

import httpx
from paddleocr import PaddleOCR  # type: ignore
from PIL import Image
from pydantic import BaseModel

from academia_mcp.files import get_workspace_dir
from academia_mcp.llm import ChatMessage, llm_acall
from academia_mcp.settings import settings

paddlex_logger = logging.getLogger("paddlex")
paddleocr_logger = logging.getLogger("paddleocr")
paddlex_logger.setLevel(logging.ERROR)
paddleocr_logger.setLevel(logging.ERROR)

DESCRIBE_PROMPTS = {
    "general": "Provide a general description of this image. Focus on the main subjects, colors, and overall scene.",
    "detailed": dedent(
        """Analyze this image in detail. Include:
        1. Main subjects and their relationships
        2. Colors, lighting, and composition
        3. Any text or symbols present
        4. Context or possible meaning
        5. Notable details or interesting elements"""
    ),
    "chess": dedent(
        """Analyze this chess position and provide a detailed description including:
        1. List of pieces on the board for both white and black
        2. Whose turn it is to move
        3. Basic evaluation of the position
        4. Any immediate tactical opportunities or threats
        5. Suggested next moves with brief explanations"""
    ),
    "text": dedent(
        """You are performing OCR and transcription.
        Extract ALL text and numbers from the image verbatim.
        - Preserve original casing, punctuation, symbols, mathematical notation, and whitespace layout when possible.
        - If layout is multi-column or tabular, reconstruct lines top-to-bottom, left-to-right; use line breaks between blocks.
        - For any uncertain or low-confidence characters, mark with a '?' and include a note.
        - After the raw extraction, provide a clean, normalized version (fixing obvious OCR artifacts) as a separate section.
        Return three sections:
        [GENERAL IMAGE DESCRIPTION]
        ...
        [RAW TRANSCRIPTION]
        ...
        [NORMALIZED TRANSCRIPTION]
        ...
        """
    ),
}


class OCRBox(BaseModel):  # type: ignore
    poly: List[List[float]]
    text: str
    score: float


class OCRSingleton:
    instance: Optional[PaddleOCR] = None
    lock: threading.Lock = threading.Lock()

    @classmethod
    def get(cls) -> PaddleOCR:
        if cls.instance is not None:
            return cls.instance
        with cls.lock:
            if cls.instance is None:
                with open(os.devnull, "w") as devnull:
                    with contextlib.redirect_stderr(devnull):
                        cls.instance = PaddleOCR(
                            use_doc_orientation_classify=False,
                            use_doc_unwarping=False,
                            use_textline_orientation=False,
                        )
        return cls.instance


async def _run_ocr(path: str) -> Dict[str, Any]:
    def _sync_ocr(path: str) -> Dict[str, Any]:
        try:
            ocr = OCRSingleton.get()
            with open(os.devnull, "w") as devnull:
                with contextlib.redirect_stderr(devnull):
                    result = ocr.predict(input=path)[0]
            rec_texts = result["rec_texts"]
            rec_scores = result["rec_scores"]
            rec_polys = result["rec_polys"]
        except Exception as e:
            return {"error": str(e)}
        items = []
        for poly, text, score in zip(rec_polys, rec_texts, rec_scores):
            items.append(OCRBox(poly=poly, text=text, score=score).model_dump())
        return {"boxes": items}

    return await asyncio.to_thread(_sync_ocr, path)


def show_image(path: str) -> Dict[str, str]:
    """
    Reads an image from the specified URL or from the current work directory.
    Always call this function at the end of the code block.
    For instance:
    ```python
    show_image("https://example.com/image.png")
    ```
    Do not print it ever, just return as the last expression.

    Args:
        path: Path to file inside current work directory or web URL
    """
    if path.startswith("http"):
        response = httpx.get(path, timeout=10)
        response.raise_for_status()
        image = Image.open(BytesIO(response.content))
    else:
        assert settings.WORKSPACE_DIR is not None, "WORKSPACE_DIR is not set"
        full_path = Path(path)
        if not full_path.exists():
            full_path = Path(get_workspace_dir()) / path
            assert full_path.exists(), f"Image file {path} does not exist"
        image = Image.open(str(full_path))
    buffer_io = BytesIO()
    image.save(buffer_io, format="PNG")
    img_bytes = buffer_io.getvalue()
    return {"image_base64": base64.b64encode(img_bytes).decode("utf-8")}


async def describe_image(
    path: str, description_type: str = "general", custom_prompt: Optional[str] = None
) -> str:
    """
    Tool to analyze and describe any image using GPT-4 Vision API.

    Returns a description of the image based on the requested type.

    Args:
        image_path (str): Path to the image file.
        description_type (str): Type of description to generate. Options:
            - "general": General description of the image
            - "detailed": Detailed analysis of the image
            - "chess": Analysis of a chess position
            - "text": Extract and describe text or numbers with an OCR pipeline.
            - "custom": Custom description based on user prompt
    """
    image_base64 = show_image(path)["image_base64"]
    assert (
        description_type in DESCRIBE_PROMPTS or description_type == "custom"
    ), f"Invalid description type: {description_type}"
    prompt = DESCRIBE_PROMPTS.get(description_type, custom_prompt)
    assert prompt and prompt.strip(), "Please provide a non-empty prompt"
    content = [
        {"type": "text", "text": prompt},
        {
            "type": "image_url",
            "image_url": {"url": f"data:image/png;base64,{image_base64}"},
        },
    ]
    model_name = settings.DESCRIBE_IMAGE_MODEL_NAME
    llm_kwargs = {}
    if description_type in {"text", "chess"}:
        llm_kwargs["temperature"] = 0.0
    response = await llm_acall(
        model_name=model_name,
        messages=[ChatMessage(role="user", content=content)],
        **llm_kwargs,
    )
    if description_type == "text":
        ocr_response = await _run_ocr(path)
        response = json.dumps(
            {
                "vlm_response": response,
                "ocr_response": ocr_response if ocr_response else [],
            }
        )
    return response



import json
import re
import shutil
import subprocess
import tempfile
from pathlib import Path
from typing import List

from pydantic import BaseModel, Field

from academia_mcp.files import (
    DEFAULT_LATEX_TEMPLATES_DIR_PATH,
    get_workspace_dir,
)
from academia_mcp.pdf import parse_pdf_file


class GetLatexTemplatesListResponse(BaseModel):  # type: ignore
    templates: List[str] = Field(description="List of available latex templates")


def get_latex_templates_list() -> GetLatexTemplatesListResponse:
    """
    Get the list of available latex templates.
    Always use one of the templates from the list.
    """
    templates = [str(path.name) for path in DEFAULT_LATEX_TEMPLATES_DIR_PATH.glob("*")]
    return GetLatexTemplatesListResponse(templates=templates)


class GetLatexTemplateResponse(BaseModel):  # type: ignore
    template: str = Field(description="The latex template")
    style: str = Field(description="The latex style")


def get_latex_template(template_name: str) -> GetLatexTemplateResponse:
    """
    Get the latex template by name.

    Args:
        template_name: The name of the latex template.
    """
    template_dir_path = DEFAULT_LATEX_TEMPLATES_DIR_PATH / template_name
    if not template_dir_path.exists():
        raise FileNotFoundError(
            f"Template {template_name} not found in {DEFAULT_LATEX_TEMPLATES_DIR_PATH}"
        )
    template_path = template_dir_path / f"{template_name}.tex"
    style_path = template_dir_path / f"{template_name}.sty"
    if not template_path.exists():
        raise FileNotFoundError(f"Template file {template_path} not found in {template_dir_path}")
    if not style_path.exists():
        raise FileNotFoundError(f"Style file {style_path} not found in {template_dir_path}")
    return GetLatexTemplateResponse(
        template=template_path.read_text(), style=style_path.read_text()
    )


def compile_latex(
    input_filename: str, output_filename: str = "output.pdf", timeout: int = 60
) -> str:
    """
    Compile a latex file.

    Returns a string with the result of the compilation.

    Args:
        input_filename: The path to the latex file.
        output_filename: The path to the output pdf file.
        timeout: The timeout for the compilation. 60 seconds by default.
    """
    input_filename_path = Path(input_filename)
    if not input_filename_path.exists():
        input_filename_path = Path(get_workspace_dir()) / input_filename
    assert input_filename_path.exists(), f"Input file {input_filename} does not exist"
    latex_code = input_filename_path.read_text(encoding="utf-8")

    if shutil.which("pdflatex") is None:
        return "pdflatex is not installed or not found in PATH."

    destination_name = (
        output_filename if output_filename.lower().endswith(".pdf") else f"{output_filename}.pdf"
    )

    try:
        with tempfile.TemporaryDirectory(
            dir=str(get_workspace_dir()), prefix="temp_latex_"
        ) as temp_dir:
            temp_dir_path = Path(temp_dir)
            tex_filename = "temp.tex"
            pdf_filename = "temp.pdf"
            tex_file_path = temp_dir_path / tex_filename
            tex_file_path.write_text(latex_code, encoding="utf-8")

            # Detect and copy local .sty packages referenced by \usepackage{...}
            # Supports optional arguments: \usepackage[opts]{pkgA,pkgB}
            try:
                package_names: set[str] = set()
                for match in re.finditer(r"\\usepackage(?:\[[^\]]*\])?\{([^}]+)\}", latex_code):
                    for name in match.group(1).split(","):
                        pkg = name.strip()
                        if pkg:
                            package_names.add(pkg)

                for pkg in package_names:
                    sty_name = f"{pkg}.sty"
                    for candidate in DEFAULT_LATEX_TEMPLATES_DIR_PATH.rglob(sty_name):
                        shutil.copyfile(candidate, temp_dir_path / sty_name)
                        break
            except Exception:
                pass

            try:
                bib_source_path = input_filename_path.parent / "references.bib"
                if bib_source_path.exists():
                    shutil.copyfile(bib_source_path, temp_dir_path / "references.bib")
            except Exception:
                pass

            try:
                subprocess.run(
                    [
                        "latexmk",
                        "-pdf",
                        "-interaction=nonstopmode",
                        "-file-line-error",
                        "-diagnostics",
                        tex_filename,
                    ],
                    cwd=str(temp_dir_path),
                    check=True,
                    capture_output=True,
                    text=True,
                    timeout=timeout,
                )
            except subprocess.TimeoutExpired:
                return f"Compilation timed out after {timeout} seconds"
            except subprocess.CalledProcessError as e:
                combined_output = (e.stdout or "") + "\n" + (e.stderr or "")
                log_path = temp_dir_path / "temp.log"
                if log_path.exists():
                    log_content = log_path.read_text(encoding="utf-8", errors="ignore")
                    combined_output = combined_output + "\n\ntemp.log content:\n" + log_content
                return f"Compilation failed. Full LaTeX output:\n{combined_output}"

            pdf_path = temp_dir_path / pdf_filename
            output_pdf_path = Path(get_workspace_dir()) / destination_name

            if pdf_path.exists():
                shutil.move(str(pdf_path), str(output_pdf_path))
                return f"Compilation successful! PDF file saved as {destination_name}"

            return (
                "Compilation completed, but PDF file was not created. Check LaTeX code for errors."
            )
    except Exception as e:
        return f"Compilation failed due to an unexpected error: {e}"


def read_pdf(pdf_path: str) -> str:
    """
    Read a PDF file to text from the file system.

    Args:
        pdf_path: The path to the pdf file in the working directory.

    Returns a JSON-serialized list of strings where each string is a page of the pdf file.
    """
    full_path = Path(get_workspace_dir()) / pdf_path
    return json.dumps(parse_pdf_file(full_path))



from academia_mcp.tools import (
    s2_get_citations,
    s2_get_references,
    s2_get_info,
    s2_search,
)


def test_s2_citations_pingpong() -> None:
    citations = s2_get_citations("2409.06820")
    assert citations.total_count >= 1
    assert "2502.18308" in str(citations.results)


def test_s2_citations_transformers() -> None:
    citations = s2_get_citations("1706.03762")
    assert citations.total_count >= 100000


def test_s2_citations_reversed() -> None:
    citations = s2_get_references("1706.03762")
    assert citations.total_count <= 100


def test_s2_citations_versions() -> None:
    citations = s2_get_citations("2409.06820v4")
    assert citations.total_count >= 1


def test_s2_get_info() -> None:
    info = s2_get_info("2409.06820")
    assert info.title is not None
    assert info.authors is not None
    assert info.external_ids is not None
    assert info.venue is not None
    assert info.citation_count is not None
    assert info.publication_date is not None
    assert info.external_ids["CorpusId"] == 272593138


def test_s2_search_base() -> None:
    result = s2_search("transformers")
    assert result.total_count >= 1
    assert "transformers" in str(result.results).lower()
    assert result.offset == 0
    assert result.returned_count == 5


def test_s2_search_offset() -> None:
    result = s2_search("transformers", offset=10)
    assert result.total_count >= 1
    assert "transformers" in str(result.results).lower()
    assert result.offset == 10
    assert result.returned_count == 5


def test_s2_search_min_citation_count() -> None:
    result = s2_search("transformers", min_citation_count=100000)
    assert result.total_count >= 2 and result.total_count <= 10


def test_s2_search_publication_date() -> None:
    result = s2_search(
        "transformers", min_citation_count=100000, publication_date="2017-01-01:2017-12-31"
    )
    assert result.total_count == 1
